from .main import create_server, start_server_async, export_data, \
                  add_model, add_models, add_ollama_model, add_ollama_models, add_zhipu_doubao


__all__ = ['create_server', 'start_server_async', 'export_data', 
           'add_model', 'add_models', 'add_ollama_model', 'add_ollama_models', 'add_zhipu_doubao']




from typing import Any
from myHttp import http
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from time import time, sleep
from _thread import start_new_thread
from .utils import endode_js, encode_engines, encode_v1_models,\
                  get_models_from_url, get_models_from_url_ollama,\
                  get_hash, handle_stream_data, handle_log_queue,\
                  generate_models_log, star_api_key
from mySecrets import hexToStr
import json
import requests
from random import randint
from queue import Queue
import os
from .logger import write_chat_completions_api, set_base_path, write_raw_api_responses,\
                   write_chat_error, write_plain_text, write_get_log, write_post_header,\
                   write_post_raw, write_config_log, add_request, extract_all_requests,\
                   extract_all_responses

__all__ = ['create_server', 'start_server_async',
           'add_model', 'add_models', 'add_ollama_model', 'add_ollama_models'
           'add_zhipu_doubao', 'export_data']

PASSWORD = ''

JS_URL = 'https://openaiapi-site.azureedge.net/public-assets/d/ddd16bc977/static/js/main.600c2350.js'

JS_CONTENT = ''
PORT = 0
DATA_DIR = ''

def init():
    global JS_CONTENT
    tmp = http(JS_URL, ToJson=False)
    if(tmp['status'] < 0 or tmp['code'] != 200):
        print('Failed to fetch JS file')
        raise Exception('Failed to fetch JS file')
    JS_CONTENT = tmp['text']
    write_plain_text('JS file loaded successfully.')
    # currently don't need to encode according model list, so just pre generate
    JS_CONTENT = endode_js(JS_CONTENT)


models = []
model_info = {} # {name: (base_url, api_key, origin_name, is_ollama:bool)}


class Request(BaseHTTPRequestHandler):
    protocol_version = "HTTP/1.1"
    def do_GET(self):
        path=self.path
        print(path)
        if(path not in ['/v1/models','/v1/engines', '/600c2350.js']
           and  not path.startswith('/v1/login/')):
            write_get_log(path, self.client_address[0], dict(self.headers), 404)
            self.send_response(404)
            self.send_header('Content-Length', 0)
            self.send_header('Connection', 'keep-alive')
            self.end_headers()
            self.wfile.flush()
            return
        if path == '/v1/models':
            write_get_log(path, self.client_address[0], dict(self.headers), 200)
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Connection', 'keep-alive')
            data = encode_v1_models(models).encode('utf-8')
            self.send_header('Content-Length', len(data))
            self.end_headers()
            self.wfile.write(data)
            self.wfile.flush()
            return
        if path == '/v1/engines':
            write_get_log(path, self.client_address[0], dict(self.headers), 200)
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Connection', 'keep-alive')
            data = encode_engines(models).encode('utf-8')
            self.send_header('Content-Length', len(data))
            self.end_headers()
            self.wfile.write(data)
            self.wfile.flush()
            return
        if path.startswith('/v1/login/'):
            path = path[len('/v1/login/'):]
            try:
                password = hexToStr(path)
            except:
                password = ''
            if(password != PASSWORD):
                print('login denied')
                write_get_log(self.path, self.client_address[0], dict(self.headers), 401)
                self.send_response(401)
                self.send_header('Access-Control-Allow-Origin', '*')
                self.send_header('Content-Type', 'application/json')
                self.send_header('Connection', 'keep-alive')
                self.send_header('Content-Length', 2)
                self.end_headers()
                self.wfile.write(b'{}')
                self.wfile.flush()
                return
            print('login approved')
            write_get_log(self.path, self.client_address[0], dict(self.headers), 200)
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-Type', 'application/json')
            self.send_header('Connection', 'keep-alive')
            self.send_header('Content-Length', 2)
            self.end_headers()
            self.wfile.write(b'{}')
            self.wfile.flush()
            return
        if path == '/600c2350.js':
            write_get_log(path, self.client_address[0], dict(self.headers), 200)
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-Type', 'application/javascript')
            self.send_header('Connection', 'keep-alive')
            js = JS_CONTENT # currently don't need to encode according model list
            js = js.encode('utf-8')
            self.send_header('Content-Length', len(js))
            self.send_header
            self.end_headers()
            self.wfile.write(js)
            self.wfile.flush()
            return
        
    def do_POST(self):
        path = self.path
        print(path)
        # return 404 for incorrect url
        if(not path.startswith('/v1/chat/completions/')):
            write_post_header(path, self.client_address[0], dict(self.headers), 'Incorrect url')
            self.send_response(404)
            self.send_header('Content-Length', 0)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Connection', 'keep-alive')
            self.end_headers()
            self.wfile.flush()
            return
        pw = path[len('/v1/chat/completions/'):]
        try:
            pw = hexToStr(pw)
        except:
            pw = ''
        # password incorrect
        if(pw != PASSWORD):
            write_post_header(path, self.client_address[0], dict(self.headers), 'Password incorrect')
            self.send_response(404)
            data = {
                "error": {
                    "message": "Not logged in, please refresh the page and input password again.", 
                    "type": "invalid_request_error",
                    "param": None,
                    "code": None
                }
            }
            data = json.dumps(data, ensure_ascii=False).encode('utf-8')
            self.send_header('Content-Length', len(data))
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-Type', 'application/json')
            self.send_header('Connection', 'keep-alive')
            self.end_headers()
            self.wfile.write(data)
            self.wfile.flush()
            return
        write_post_header(path, self.client_address[0], dict(self.headers), 'Login success')
        body = self.rfile.read(int(self.headers['Content-Length']))
        stream_id = get_hash(str(time()) + str(randint(0, 10000000000)) + str(body))
        full_id = stream_id
        stream_id = stream_id[-12:]
        write_post_raw(path, self.client_address[0], dict(self.headers), body, stream_id)
        body = json.loads(body)
        model_name = body['model']
        # model not found
        if(model_name not in models):
            write_plain_text(f'{stream_id}  Model {model_name} not found in {models}.')
            print(f'Model {model_name} not found.')
            self.send_response(404)
            data = {
                "error": {
                    "message": f"No model named {model_name}.", 
                    "type": "invalid_request_error",
                    "param": None,
                    "code": None
                }
            }
            data = json.dumps(data, ensure_ascii=False).encode('utf-8')
            self.send_header('Content-Length', len(data))
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-Type', 'application/json')
            self.send_header('Connection', 'keep-alive')
            self.end_headers()
            self.wfile.write(data)
            self.wfile.flush()
            return
        print(body)
        # set url
        base_url = model_info[model_name][0]
        api_key = model_info[model_name][1]
        origin_name = model_info[model_name][2]
        is_ollama = model_info[model_name][3]
        if(is_ollama):
            url = base_url + '/v1/chat/completions'
        else:
            url = base_url + '/chat/completions'
        body['model'] = origin_name
        # update json data, change to old version
        for i in range(len(body['messages'])-1, -1, -1):
            if(len(body['messages'][i]['content']) == 0):
                body['messages'].pop(i)
                continue
            if(len(body['messages'][i]['content'])>1 or body['messages'][i]['content'][0]['type'] != 'text'):
                write_plain_text(f'{stream_id}  Unsupported type of message received.')
                print('Unsupported type of message received.')
                self.send_response(404)
                data = {
                    "error": {
                        "message": "Image and other type of message is not supported.", 
                        "type": "invalid_request_error",
                        "param": None,
                        "code": None
                    }
                }
                data = json.dumps(data, ensure_ascii=False).encode('utf-8')
                self.send_header('Content-Length', len(data))
                self.send_header('Access-Control-Allow-Origin', '*')
                self.send_header('Content-Type', 'application/json')
                self.send_header('Connection', 'keep-alive')
                self.end_headers()
                self.wfile.write(data)
                self.wfile.flush()
                return
            body['messages'][i]['content'] = body['messages'][i]['content'][0]['text']
        write_chat_completions_api(stream_id, json.dumps(body,ensure_ascii=False), model_name, origin_name, url)
        body = json.dumps(body, ensure_ascii=False).encode('utf-8')
        header = {
            'Authorization': f'Bearer {api_key}',
            'Content-Type': 'application/json',
            'Content-Length': str(len(body)),
        }
        add_request(full_id, body.decode('utf-8'), model_name, origin_name)
        resp = requests.post(url, headers=header, data=body, stream=True)
        # success, in transfer-encoding chunked mode
        if(resp.status_code == 200 and resp.headers.get('Transfer-Encoding') == 'chunked'):
            # use stream here
            client_queue = Queue() # sent to client
            log_queue = Queue() # for logs
            start_new_thread(handle_stream_data, (self, client_queue))
            start_new_thread(handle_log_queue, (log_queue, stream_id, full_id))
            for chunk in resp.iter_content(chunk_size=None):
                client_queue.put(chunk)
                log_queue.put(chunk)
            print("stream ended")
            client_queue.put(False)
            log_queue.put(False)
            return
        else:
            # error handling, don't use stream
            s = resp.status_code
            data = resp.content
            # special case for cohere, https://github.com/missuo/cohere2openai
            # it will not have transfer-encoding chunked if response is too short.
            if(b'data: [DONE]' in data):
                # write_raw_api_responses(stream_id, data, 0)
                # write_raw_api_responses(stream_id, "END OF RESPONSE", 1)
                # self.send_response(200)
                # self.send_header('Transfer-Encoding', 'chunked')
                # self.send_header('Access-Control-Allow-Origin', '*')
                # self.send_header('Content-Type', 'text/event-stream; charset=utf-8')
                # self.send_header('Connection', 'keep-alive')
                # self.end_headers()
                # self.wfile.write(f'{len(data):X}'.encode('utf-8'))
                # self.wfile.write(b'\r\n')
                # self.wfile.write(data)
                # self.wfile.write(b'\r\n')
                # self.wfile.write(b'0\r\n\r\n')
                # self.wfile.flush()
                client_queue = Queue() # sent to client
                log_queue = Queue() # for logs
                start_new_thread(handle_stream_data, (self, client_queue))
                start_new_thread(handle_log_queue, (log_queue, stream_id, full_id))
                log_queue.put(data)
                client_queue.put(data)
                print("stream ended")
                client_queue.put(False)
                log_queue.put(False)
                return
            # put http status and error msg from that service in error message, as detailed as possible
            # also give some hint in some case
            try:
                data = data.decode('utf-8')
            except:
                data = str(data)[2:-1]
            write_chat_error(stream_id, data, s)
            msg = f'Error: http status {s}. Error message from {url}: {data}'
            # special case for cohere
            if(s == 200):
                msg += ' This means an error occurred, but that service doesn\'t provide error message, and \
                    has status 200. But this is an error, may due to invalid api key.'
            # for zhipu
            if('"code":"1210"' in msg):
                msg += ' 这个错误来自智谱AI, 原因大概率是 temperature 和 top_p 设置不当。请注意, 智谱AI要求 temperature 和 top_p 都是 严格大于0, 严格小于1。'
            data = {
                "error": {
                    "message": msg, 
                    "type": "invalid_request_error",
                    "param": None,
                    "code": None
                }
            }
            self.send_response(404)
            data = json.dumps(data, ensure_ascii=False).encode('utf-8')
            self.send_header('Content-Length', len(data))
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-Type', 'application/json')
            self.send_header('Connection', 'keep-alive')
            self.end_headers()
            self.wfile.write(data)
            self.wfile.flush()
            return
        
    
    def do_OPTIONS(self):
        self.send_response(200)
        self.send_header('Content-Length', 0)
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Headers', 'openai-organization,content-type,authorization,openai-project')
        self.end_headers()
        self.wfile.flush()
    
    def log_message(self, *args) -> None:
        pass
    
    def handle(self):
        try:
            super().handle()
        except (BrokenPipeError, ConnectionResetError):
            pass


init()


def create_server(port:int, password:str, data_dir:str='./playground_logs') -> None:
    '''
    Create the playground server. Just create, will not start. You need to add models to it and then call `start_server_async` later.
    
    Arguments:
    
    1. password: the password the needs to be input in the browser, for authorization
    2. data_dir: the directory to store the database and logs. Will create if not exists.
    '''
    global PORT, PASSWORD
    if(port <=0 or port >=65536):
        raise ValueError('Invalid port number: ' + str(port))
    # PORT = port
    if(len(password) == 0):
        raise TypeError('Password cannot be empty.')
    PASSWORD = password
    if(data_dir.endswith('/')):
        data_dir = data_dir[:-1]
    os.makedirs(data_dir, exist_ok=True)
    set_base_path(data_dir)
    PORT = port
    write_config_log(f"Create server, with port {PORT} and password {PASSWORD}. Latest config: " + generate_models_log(model_info))

def add_model(base_url:str, api_key:str, model_name:str, new_name: str=None) -> None:
    '''
    Add a model to the playground, need openai format API. Need to implement both `<base_url>/models` and `<base_url>/chat/completions`
    
    Arguments:
    
    1. base_url: should include `/v1` if it's openai, start with `http://` or `https://`. Either have or not have the last `/` is OK. E.g. for openai, it should be `https://api.openai.com/v1`. It will access `<base_url>/models` and `<base_url>/chat/completions`
    2. model_name: need to be exactly same as the name in your service
    3. new_name: the name you want to show in the playground, if not provided, will be same as model_name
    '''
    write_config_log(f"add_model: base_url: {base_url}, api_key: {star_api_key(api_key)}, model_name: {model_name}, new_name: {new_name}")
    if(not base_url.startswith('http://') and not base_url.startswith('https://')):
        raise ValueError('Invalid url. Url should start with http:// or https://')
    if(base_url[-1] == '/'):
        base_url = base_url[:-1]
    if(model_name == '' or new_name == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name == None):
        new_name = model_name
    if(model_name.replace(' ', '').replace('\n', '') == '' or new_name.replace(' ', '').replace('\n', '') == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name in models):
        raise ValueError(f'Model name {new_name} already exists.')
    all_models = get_models_from_url(base_url, api_key)
    if(model_name not in all_models):
        raise ValueError(f'No model called {model_name} from {base_url}. Available models: {all_models}')
    models.append(new_name)
    model_info[new_name] = (base_url, api_key, model_name, False)
    print(f'Model {new_name} added successfully.')
    write_config_log("Model added, latest config: " + generate_models_log(model_info))


def add_models(base_url:str, api_key:str, models_:list[str] = [], prefix:str='', postfix:str='') -> None:
    '''
    Add a list of models to the playground, need openai format API. Need to implement both `<base_url>/models` and `<base_url>/chat/completions`. If models_ is empty, will add all available models.
    
    Arguments:
    
    1. base_url: should include `/v1` if it's openai, start with `http://` or `https://`. Either have or not have the last `/` is OK. E.g. for openai, it should be `https://api.openai.com/v1`. It will access `<base_url>/models` and `<base_url>/chat/completions`
    2. models_: a list of model names you want to add, model name should be exactly same as the name in your service. If it is an empty list, will add all available models.
    3. prefix and postfix: the name shown in the playground will be `prefix + model_name + postfix`
    '''
    write_config_log(f"add_models: base_url: {base_url}, api_key: {star_api_key(api_key)}, models_: {models_}, prefix: {prefix}, postfix: {postfix}")
    if(not base_url.startswith('http://') and not base_url.startswith('https://')):
        raise ValueError('Invalid url. Url should start with http:// or https://')
    if(base_url[-1] == '/'):
        base_url = base_url[:-1]
    # first check the model name legality
    if(len(set(models_)) != len(models_)):
        raise ValueError('Have duplicate model names.')
    for model_name in models_:
        if(model_name == ''):
            raise ValueError('Model name cannot be empty.')
        if(model_name.replace(' ', '').replace('\n', '') == ''):
            raise ValueError('Model name cannot be empty.')
        if((prefix + model_name + postfix) in models):
            raise ValueError(f'Model name {prefix + model_name + postfix} already exists.')
    all_models = get_models_from_url(base_url, api_key)
    if(len(models_) ==0):
        models_ = all_models
    for model_name in models_:
        if((prefix + model_name + postfix) in models):
            raise ValueError(f'Model name {prefix + model_name + postfix} already exists.')
    added_models = []
    missing_models = []
    for m in models_:
        if(m in all_models):
            added_models.append(prefix + m + postfix)
            models.append(prefix + m + postfix)
            model_info[prefix + m + postfix] = (base_url, api_key, m, False)
        else:
            missing_models.append(m)
    write_config_log("Model added, latest config: " + generate_models_log(model_info))
    if(len(missing_models) ==0):
        print(f'All models added successfully. {added_models}')
        return
    if(len(added_models) > 0):
        print(f'These models added successfully: {added_models}')
    print(f'These models not found: {missing_models} from {base_url}. Available models: {all_models}')
    raise Exception()


def add_ollama_model(base_url:str, api_key:str, model_name:str, new_name: str=None) -> None:
    '''
    For ollama, if you don't know what ollama is, just ignore this function.
    
    Mostly same as `add_model`.
    
    For base_url, on default should end with :11434, no any other path. Just ends with port number (if you change the default port, port can be different)
    
    For api_key, it's not checked, but you must provide, even an empty str is OK.
    '''
    write_config_log(f"add_ollama_model: base_url: {base_url}, api_key: {star_api_key(api_key)}, model_name: {model_name}, new_name: {new_name}")
    if(not base_url.startswith('http://') and not base_url.startswith('https://')):
        raise ValueError('Invalid url. Url should start with http:// or https://')
    if(base_url[-1] == '/'):
        base_url = base_url[:-1]
    if(model_name == '' or new_name == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name == None):
        new_name = model_name
    if(model_name.replace(' ', '').replace('\n', '') == '' or new_name.replace(' ', '').replace('\n', '') == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name in models):
        raise ValueError(f'Model name {new_name} already exists.')
    all_models = get_models_from_url_ollama(base_url, api_key)
    if(model_name not in all_models):
        raise ValueError(f'No model called {model_name} from {base_url}. Available models: {all_models}')
    models.append(new_name)
    model_info[new_name] = (base_url, api_key, model_name, True)
    print(f'Model {new_name} added successfully.')
    write_config_log("Model added, latest config: " + generate_models_log(model_info))


def add_ollama_models(base_url:str, api_key:str, models_:list[str] = [], prefix:str='', postfix:str='') -> None:
    '''
    For ollama, if you don't know what ollama is, just ignore this function.
    
    Mostly same as `add_models`.
    
    For base_url, on default should end with :11434, no any other path. Just ends with port number (if you change the default port, port can be different)
    
    For api_key, it's not checked, but you must provide, even an empty str is OK.
    '''
    write_config_log(f"add_ollama_models: base_url: {base_url}, api_key: {star_api_key(api_key)}, models_: {models_}, prefix: {prefix}, postfix: {postfix}")
    if(not base_url.startswith('http://') and not base_url.startswith('https://')):
        raise ValueError('Invalid url. Url should start with http:// or https://')
    if(base_url[-1] == '/'):
        base_url = base_url[:-1]
    # first check the model name legality
    if(len(set(models_)) != len(models_)):
        raise ValueError('Have duplicate model names.')
    for model_name in models_:
        if(model_name == ''):
            raise ValueError('Model name cannot be empty.')
        if(model_name.replace(' ', '').replace('\n', '') == ''):
            raise ValueError('Model name cannot be empty.')
        if((prefix + model_name + postfix) in models):
            raise ValueError(f'Model name {prefix + model_name + postfix} already exists.')
    all_models = get_models_from_url_ollama(base_url, api_key)
    if(len(models_) ==0):
        models_ = all_models
    for model_name in models_:
        if((prefix + model_name + postfix) in models):
            raise ValueError(f'Model name {prefix + model_name + postfix} already exists.')
    added_models = []
    missing_models = []
    for m in models_:
        if(m in all_models):
            added_models.append(prefix + m + postfix)
            models.append(prefix + m + postfix)
            model_info[prefix + m + postfix] = (base_url, api_key, m, True)
        else:
            missing_models.append(m)
    write_config_log("Model added, latest config: " + generate_models_log(model_info))
    if(len(missing_models) ==0):
        print(f'All models added successfully. {added_models}')
        return
    if(len(added_models) > 0):
        print(f'These models added successfully: {added_models}')
    print(f'These models not found: {missing_models} from {base_url}. Available models: {all_models}')
    raise Exception()


def add_zhipu_doubao(base_url: str, api_key:str, model_name:str, new_name: str=None) -> None:
    '''
    For non-Chinese users, just ignore this function.
    
    针对没有 `/v1/models`、只有 `/v1/chat/completions` 的API (目前已知的有智谱AI和豆包)
    
    这个函数会直接把模型添加进去, 不会检查 可用的模型和 API key 是否正确
    
    其它和 `add_model` 相同, 这个函数没有 一次添加多个模型的版本
    '''
    write_config_log(f"add_zhipu_doubao: base_url: {base_url}, api_key: {star_api_key(api_key)}, model_name: {model_name}, new_name: {new_name}")
    if(not base_url.startswith('http://') and not base_url.startswith('https://')):
        raise ValueError('Invalid url. Url should start with http:// or https://')
    if(base_url[-1] == '/'):
        base_url = base_url[:-1]
    if(model_name == '' or new_name == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name == None):
        new_name = model_name
    if(model_name.replace(' ', '').replace('\n', '') == '' or new_name.replace(' ', '').replace('\n', '') == ''):
        raise ValueError('Model name cannot be empty.')
    if(new_name in models):
        raise ValueError(f'Model name {new_name} already exists.')
    models.append(new_name)
    model_info[new_name] = (base_url, api_key, model_name, False)
    print(f'Model {new_name} added, but its availability and correctness of api key is not tested.')
    write_config_log("Model added, latest config: " + generate_models_log(model_info))


def start_server_async() -> None:
    '''
    Start the server, in async mode. (return immediately, run in background, the code after this can be executed)
    
    This function should be called after `create_server`.
    '''
    if (PORT == 0):
        raise Exception("Please first call create_server.")
    server = ThreadingHTTPServer(('0.0.0.0', PORT), Request)
    write_config_log("Start the server, latest config: " + generate_models_log(model_info))
    start_new_thread(server.serve_forever, ())
    print(f'Server started, latest models: {generate_models_log(model_info)}')


def export_data():
    '''
    Export the data of history requests and responses.
    '''
    all_requests = extract_all_requests()
    all_responses = extract_all_responses()
    all = []
    responses_dict = {}
    for r in all_responses:
        responses_dict[r[0]] = r
    for r in all_requests:
        record = {}
        request_id = r[0]
        request_data = r[1]
        request_time = r[2]
        has_response = r[3]
        user_model_name = r[4]
        origin_model_name = r[5]
        record['id'] = request_id
        record['request_data'] = json.loads(request_data) # it must be json
        record['request_time_us'] = request_time
        record['has_response'] = has_response
        record['user_model_name'] = user_model_name
        record['origin_model_name'] = origin_model_name
        if(has_response):
            response = responses_dict[request_id]
            response_data = response[1]
            first_response_time = response[2]
            end_time = response[3]
            ended_successfully = response[4]
            input_tokens = response[5]
            output_tokens = response[6]
            record['response_data'] = response_data
            record['first_response_time_us'] = first_response_time
            record['end_time_us'] = end_time
            record['ended_successfully'] = ended_successfully
            record['input_tokens'] = input_tokens
            record['output_tokens'] = output_tokens
        all.append(record)
    return all
            
        


if __name__ == '__main__':
    # from keys import OPENAI_API_KEY, COHERE_API_KEY, OLLAMA_API_KEY, ZHIPU_API_KEY, KIMI_API_KEY, DOUBAO_API_KEY
    create_server(9025, 'jtc1246')
    print(export_data())
    add_model('http://jtc1246.com:9002/v1/',COHERE_API_KEY,'command-r-plus','cohere')
    # add_models('https://api.openai.com/v1/', OPENAI_API_KEY, ['gpt-3.5-turbo','gpt-4'], prefix='openai-')
    add_model('https://api.openai.com/v1/', OPENAI_API_KEY, 'gpt-4-turbo-2024-04-09', 'openai-gpt-4')
    add_model('https://api.openai.com/v1/', OPENAI_API_KEY, 'gpt-4o-2024-05-13', 'openai-gpt-4o')
    add_model('https://api.openai.com/v1/', OPENAI_API_KEY, 'gpt-3.5-turbo-0125', 'openai-gpt-3.5')
    add_ollama_models('http://127.0.0.1:11434', 'ollama')
    add_models('https://api.moonshot.cn/v1', KIMI_API_KEY)
    add_zhipu_doubao('https://open.bigmodel.cn/api/paas/v4/', ZHIPU_API_KEY, 'glm-4')
    add_zhipu_doubao('https://ark.cn-beijing.volces.com/api/v3', DOUBAO_API_KEY, 'ep-20240709181337-fmg27', 'doubao-pro-128k-240628')
    start_server_async()
    while True:
        sleep(10)



import json
from myHttp import http
from hashlib import sha256
import os
from queue import Queue
from .logger import write_raw_api_responses, write_config_log, write_plain_response,\
                   add_response, update_response, set_token_usage, write_plain_text

__all__ = ['endode_js', 'encode_engines', 'encode_v1_models',
           'get_models_from_url', 'get_models_from_url_ollama',
           'get_hash', "handle_stream_data"]


def endode_js(js: str):
    '''
    Modify the js file that it thinks all model names are valid.
    '''
    origin_js = js
    found_replacer = True
    if (js.find('/^gpt-3\.5-turbo(?!-instruct)(?!-base)|^gpt-4(?!-base)|^gpt-dv/') >= 0):
        # print('found')
        write_plain_text('REPLACER_FOUND  /^gpt-3\.5-turbo(?!-instruct)(?!-base)|^gpt-4(?!-base)|^gpt-dv/')
        pass
    else:
        # print('not found')
        write_plain_text('REPLACER_NOT_FOUND  /^gpt-3\.5-turbo(?!-instruct)(?!-base)|^gpt-4(?!-base)|^gpt-dv/')
        found_replacer = False
    js = js.replace('/^gpt-3\.5-turbo(?!-instruct)(?!-base)|^gpt-4(?!-base)|^gpt-dv/', '/^.*$/')
    if (js.find('/^gpt-4[a-z]?-(?!vision)(?!base).*/') >= 0):
        # print('found')
        write_plain_text('REPLACER_FOUND  /^gpt-4[a-z]?-(?!vision)(?!base).*/')
        pass
    else:
        # print('not found')
        write_plain_text('REPLACER_NOT_FOUND  /^gpt-4[a-z]?-(?!vision)(?!base).*/')
        found_replacer = False
    if(found_replacer == False):
        write_plain_text('Warning: Some text that needed to replace is not found in the current js file (because it has changed from origin one). The website might work normally, might not work. If it doesn\'t work, this may be a reason.')
        print('Warning: Some text that needed to replace is not found in the current js file (because it has changed from origin one). The website might work normally, might not work. If it doesn\'t work, this may be a reason.')
    js = js.replace('/^gpt-4[a-z]?-(?!vision)(?!base).*/', '/^.*$/')
    js = js.replace('||2049', '||4097')
    js = js.replace('"/v1/chat/completions"', '"/v1/chat/completions/"+jtc_password')
    file_path = os.path.abspath(__file__)
    file_path = os.path.dirname(file_path) + '/append.js'
    with open(file_path, 'r') as f:
        tmp = f.read()
        js += tmp
    return js


def encode_engines(models: list[str]) -> str:
    data = {"object": "list"}
    l = []
    for m in models:
        l.append({
            "object": "engine",
            "id": m,
            "ready": True,
            "owner": "system",
            "permissions": None,
            "created": None
        })
    data['data'] = l
    return json.dumps(data, ensure_ascii=False)


def encode_v1_models(models: list[str]) -> str:
    data = {"object": "list"}
    l = []
    for m in models:
        l.append({
            "object": "model",
            "id": m,
            "owned_by": "system",
            "created": 1715367049
        })
    data['data'] = l
    # print(json.dumps(data))
    return json.dumps(data, ensure_ascii=False)


def extract_models(data: str) -> list[str]:
    '''
    extract the models from json, if it's from openai, will delete like whisper and tts
    '''
    block_list = ['whisper-','tts-','dall-e','embedding-','babbage-','davinci-','ada-']
    
    try:
        if (type(data) == str):
            data = json.loads(data)
        data2 = data['data']
        results = []
        has_gpt = False
        for info in data2:
            name = info['id']
            assert (type(name) == str)
            if (str.find(name, 'gpt') >= 0):
                has_gpt = True
            results.append(name)
        if(has_gpt):
            l = len(results)
            for i in range(l-1, -1, -1):
                that_name = results[i]
                need_delete = False
                for blocked in block_list:
                    if(str.find(that_name, blocked) >= 0):
                        need_delete = True
                        break
                if(need_delete):
                    results.pop(i)
        return results
    except:
        print(data)
        raise TypeError('Invalid response from server')


def extract_models_ollama(data: str) -> list[str]:
    '''
    extract models from json, for ollama
    '''
    try:
        if (type(data) == str):
            data = json.loads(data)
        data2 = data['models']
        results = []
        has_gpt = False
        for info in data2:
            name = info['model']
            assert (type(name) == str)
            if (str.find(name, 'gpt') >= 0):
                has_gpt = True
            results.append(name)
        return results
    except:
        print(data)
        raise TypeError('Invalid response from server')


def get_models_from_url(base_url: str, api_key: str):
    '''
    url: not ends with /, usually should ends with /v1
    
    will not handle whether api key is correct here, if server doesn't do verification on /v1/models
    '''
    # this comment is the latest: currently designed to provide separate functions for zhipu 
    #     and doubao, which don't have models api, just think all models are available
    # # for zhipu, because it doesn't have models api
    # if('zhipu' in base_url or 'bigmodel.cn' in base_url):
    #     return ['glm-4-0520', 'glm-4', 'glm-4-AirX', 'glm-4-Air', 'glm-4-Flash', 'chatglm-3', 'glm-3-turbo']
    url = base_url + '/models'
    header = {
        'Authorization': f'Bearer {api_key}',
        'Content-Type': 'application/json',
    }
    resp = http(url, Header=header)
    if (resp['status'] < 0):
        raise ConnectionError(f"Can't connect to {base_url}")
    if (resp['status'] > 0):
        write_config_log(f"Error in getting models from {url},  " + str(resp))
        raise ConnectionError(f"Invalid response from server, " + str(resp['extra']))
    if(resp['code'] != 200):
        print(f'Error: status code {resp["code"]}')
        print(resp['text'])
        write_config_log(f"Error in getting models from {url},  " + str(resp))
        raise Exception("Invalid api key, or other server error.")
    write_config_log(f"Models from {url},  " + json.dumps(resp['text'], ensure_ascii=False))
    return extract_models(resp['text'])


def get_models_from_url_ollama(base_url: str, api_key: str):
    '''
    url: on default should end with :11434, no any other path
    
    just ends with port number (if you change the default port, port can be different)
    '''
    url = base_url + '/api/tags'
    header = {
        'Authorization': f'Bearer {api_key}',
        'Content-Type': 'application/json',
    }
    resp = http(url, Header=header)
    if (resp['status'] < 0):
        raise ConnectionError(f"Can't connect to {base_url}")
    if (resp['status'] > 0):
        write_config_log(f"Error in getting models from {url},  " + str(resp))
        raise ConnectionError(f"Invalid response from server, " + str(resp['extra']))
    if(resp['code'] != 200):
        print(f'Error: status code {resp["code"]}')
        print(resp['text'])
        write_config_log(f"Error in getting models from {url},  " + str(resp))
        raise Exception("Invalid api key, or other server error.")
    write_config_log(f"Models from {url},  " + json.dumps(resp['text'], ensure_ascii=False))
    return extract_models_ollama(resp['text'])


def get_hash(s: str) -> str:
    return sha256(s.encode()).hexdigest()


def handle_stream_data(request_obj, data_queue:Queue):
    request_obj.send_response(200)
    request_obj.send_header('Access-Control-Allow-Origin', '*')
    request_obj.send_header('Content-Type', 'text/event-stream; charset=utf-8')
    request_obj.send_header('Transfer-Encoding', 'chunked')
    request_obj.send_header('Connection', 'keep-alive')
    request_obj.end_headers()
    try:
        while True:
            data = data_queue.get()
            if(data == False):
                break
            request_obj.wfile.write(f"{len(data):x}".encode('utf-8'))
            request_obj.wfile.write(b'\r\n')
            request_obj.wfile.write(data)
            request_obj.wfile.write(b'\r\n')
        request_obj.wfile.write(b'0\r\n\r\n')
        request_obj.wfile.flush()
    except:
        pass


def construct_response(data: bytes):
    while(len(data) > 0 and data[-1] == b'\n'):
        data = data[:-1]
    if(len(data) == 0):
        return ''
    parts = data.split(b'\n\n')
    result = ''
    for p in parts:
        p = p[6:]
        try:
            p = p.decode('utf-8')
            p = json.loads(p)
            result += p['choices'][0]['delta']['content']
        except:
            continue
    return result


def get_usage(data: bytes):
    data = data.decode('utf-8')[:-2] # remove the last 2 \n
    parts = data.split('\n\n')
    last_msg = parts[-2]
    last_msg = json.loads(last_msg[6:])
    try:
        in_tokens = last_msg['usage']['prompt_tokens']
        out_tokens = last_msg['usage']['completion_tokens']
        return (in_tokens, out_tokens)
    except:
        return (None, None)


def handle_log_queue(log_queue:Queue, stream_id:str, full_id:str):
    index = 0
    data_till_now = b''
    latest_resp = ''
    added = False
    while True:
        data = log_queue.get()
        if (data == False):
            write_raw_api_responses(stream_id, "END OF RESPONSE", index)
            in_tokens, out_tokens = get_usage(data_till_now)
            print(f'In tokens: {in_tokens}, out tokens: {out_tokens}')
            write_plain_response(stream_id, str([latest_resp]) + ' FINISHED  ' + f'In: {in_tokens}, Out: {out_tokens}', index)
            set_token_usage(full_id, in_tokens, out_tokens)
            break
        data_till_now += data
        latest_resp = construct_response(data_till_now)
        write_raw_api_responses(stream_id, data, index)
        write_plain_response(stream_id, str([latest_resp]), index)
        if(added == False):
            added = True
            add_response(full_id, latest_resp)
        else:
            update_response(full_id, latest_resp)
        index += 1


def generate_models_log(model_info:dict):
    tmp = json.dumps(model_info, ensure_ascii=False)
    model_info = json.loads(tmp)
    for m in model_info:
        model_info[m] = list(model_info[m])
        model_info[m][1] = '' # delete api key
    return json.dumps(model_info, ensure_ascii=False)


def star_api_key(api_key:str) -> str:
    l = len(api_key)
    s = l // 6
    e = l - s
    return api_key[:s] + '*' * (l - s * 2) + api_key[e:]


if __name__ == '__main__':
    # from keys import OPENAI_API_KEY, COHERE_API_KEY, OLLAMA_API_KEY
    print(get_models_from_url('https://api.openai.com/v1', OPENAI_API_KEY))
    print(get_models_from_url('http://jtc1246.com:9002/v1', COHERE_API_KEY))
    print(get_models_from_url_ollama('http://127.0.0.1:11434', 'ollama'))
    


import os
from queue import Queue
from time import time
from _thread import start_new_thread
import json
import sqlite3
from threading import Lock


LOG_BASE_PATH = ''
log_file = None
database = None


log_queue = Queue()
database_lock = Lock()


__all__ = ['set_base_path', 'write_raw_api_responses', 'write_chat_completions_api',
           'write_chat_error', 'write_plain_text', 'write_config_log', 'write_get_log', 
           'write_post_header', 'write_post_raw', 'write_plain_response',
           'add_request', 'extract_all_requests',
           'extract_all_responses', 'add_response', 'update_response', 'set_token_usage']


def set_base_path(p: str):
    global LOG_BASE_PATH, log_file, database
    LOG_BASE_PATH = p
    log_file = open(p + '/playground_logs.txt', 'a')
    log_file.write('\n\n')
    start_new_thread(write_queue, (log_queue, log_file))
    log_queue.put('Server started')
    database = sqlite3.connect(p + '/playground_database.sqlite3', check_same_thread=False)
    # time is in microsecond, all times after this are in microsecond
    database.execute('''
        Create Table If Not Exists All_Requests (
            id Char(64) Primary Key,
            data Text Not Null,
            time Integer Not Null,
            has_response Boolean Not Null,
            user_model_name Text Not Null,
            origin_model_name Text Not Null
        );
                     ''')
    # start time is the time of receiving first response
    database.execute('''
        Create Table If Not Exists All_Responses (
            id Char(64) Primary Key,
            data Text Not Null,
            start_time Integer Not Null,
            last_update_time Integer Not Null,
            ended Boolean Not Null,
            in_tokens Integer,
            out_tokens Integer
        );
                     ''')
    database.commit()


def write_raw_api_responses(stream_id: str, data: bytes, index: int):
    try:
        data = data.decode('utf-8')
        data = str([data])
    except:
        data = str(data)
    index = str(index)
    index = index + ' ' * (5 - len(index))
    content = 'RAW_RESP' + ',  ' + stream_id + '  ' + index + ' ' + data
    log_queue.put(content)


def write_chat_completions_api(stream_id: str, data: str, request_model: str, used_model: str, base_url: str):
    # later part contains more information
    if (len(request_model) > 30):
        request_model = request_model[-30:]
    if (len(used_model) > 30):
        used_model = used_model[-30:]
    request_model = request_model + ' ' * (30 - len(request_model))
    used_model = used_model + ' ' * (30 - len(used_model))
    content = 'CHAT_API' + ',  ' + stream_id + '  R:' + request_model + '  U:' + used_model + '  ' + data + '  ' + base_url
    log_queue.put(content)


def write_chat_error(stream_id: str, data: str, status: int):
    status = str(status)
    status = status + ' ' * (3 - len(status))
    content = 'CHAT_ERR' + ',  ' + stream_id + '  ' + status + '  ' + data
    log_queue.put(content) 


def write_plain_text(data: str):
    content = 'PLN_TEXT' + ',  ' + data
    log_queue.put(content)


def write_config_log(data: str):
    content = 'CONFIG__' + ',  ' + data
    log_queue.put(content)


def write_get_log(path, ip, header, status):
    status = str(status)
    status = status + ' ' * (3 - len(status))
    ip = ip + ' ' * (15 - len(ip))
    path = path + ' ' * (50 - len(path))
    content = 'HTTP_GET' + ',  ' + path + '  ' + status + '  ' + ip + '  ' + json.dumps(header, ensure_ascii=False)
    log_queue.put(content)


def write_post_header(path: str, ip, header, desc):
    '''
    This is only for wrong path or password, wrong model name doesn't belong to this
    '''
    desc = desc + ' ' * (20 - len(desc))
    ip = ip + ' ' * (15 - len(ip))
    path = path + ' ' * (50 - len(path))
    content = 'POST_HDR,  ' + desc + '  ' + path + '  ' + ip + '  ' + json.dumps(header, ensure_ascii=False)
    log_queue.put(content)


def write_post_raw(path: str, ip: str, header: str, data: bytes, stream_id: str):
    try:
        data = data.decode('utf-8')
        data = str([data])
    except:
        data = str(data)
    ip = ip + ' ' * (15 - len(ip))
    path = path + ' ' * (50 - len(path))
    content = 'POST_RAW' + ',  ' + stream_id + '  ' + path + '  ' + ip + '  ' + json.dumps(header, ensure_ascii=False) + '  ' + data # bytes
    log_queue.put(content)


def write_plain_response(stream_id:str, data: str, index: int):
    index = str(index)
    index = index + ' ' * (5 - len(index))
    content = "PURE_RSP,  " + stream_id + '  ' + index + ' ' + data
    log_queue.put(content)


# def write


def write_queue(q: Queue, file):
    while True:
        content = q.get()
        t = "{:.3f}".format(time() * 1000)
        content = t + ':  ' + content + '\n'
        file.write(content)
        file.flush()


def add_request(full_id: str, data, user_model_name: str, origin_model_name: str):
    if(type(data) != str):
        data = json.dumps(data, ensure_ascii=False)
    t = int(round(time()*1000000))
    database_lock.acquire()
    database.execute('''
        Insert Into All_Requests (id, data, time, has_response, user_model_name, origin_model_name)
        Values (?, ?, ?, ?, ?, ?);
                     ''', (full_id, data, t, False, user_model_name, origin_model_name))
    database.commit()
    database_lock.release()


def set_has_response(full_id: str):
    database.execute('''
        Update All_Requests
        Set has_response = True
        Where id = ?;
                     ''', (full_id,))
    # database.commit() # don't need to commit here, need to commit once with add_response


def extract_all_requests():
    cursor = database.execute('''
        Select *
        From All_Requests;
                             ''')
    result = cursor.fetchall()
    my_results = []
    for r in result:
        my_results.append(list(r))
        my_results[-1][3] = bool(my_results[-1][3])
    return my_results


def extract_all_responses():
    cursor = database.execute('''
        Select *
        From All_Responses;
                             ''')
    result = cursor.fetchall()
    my_results = []
    for r in result:
        my_results.append(list(r))
        my_results[-1][4] = bool(my_results[-1][4])
    return my_results


def add_response(full_id:str, data:str):
    t = int(round(time()*1000000))
    database_lock.acquire()
    database.execute('''
        Insert Into All_Responses (id, data, start_time, last_update_time, ended)
        Values (?, ?, ?, ?, ?);
                     ''', (full_id, data, t, t, False))
    set_has_response(full_id)
    database.commit()
    database_lock.release()
    

def update_response(full_id: str, data: str):
    t = int(round(time()*1000000))
    database_lock.acquire()
    database.execute('''
        Update All_Responses
        Set data = ?, last_update_time = ?
        Where id = ?;
                     ''', (data, t, full_id))
    database.commit()
    database_lock.release()


def set_token_usage(full_id: str, in_tokens: int = None, out_tokens: int = None):
    t = int(round(time()*1000000))
    database_lock.acquire()
    database.execute('''
        Update All_Responses
        Set in_tokens = ?, out_tokens = ?, last_update_time = ?, ended = True
        Where id = ?;
                     ''', (in_tokens, out_tokens, t, full_id))
    database.commit()
    database_lock.release()


if __name__ == '__main__':
    database = sqlite3.connect('playground_logs/playground_database.sqlite3', check_same_thread=False)
    # time 是 微秒, 之后所有的时间都是微秒
    database.execute('''
        Create Table If Not Exists All_Requests (
            id Char(64) Primary Key,
            data Text Not Null,
            time Integer Not Null,
            has_response Boolean Not Null,
            user_model_name Text Not Null,
            origin_model_name Text Not Null
        );
                     ''')
    database.execute('''
        Create Table If Not Exists All_Responses (
            id Char(64) Primary Key,
            data Text Not Null,
            start_time Integer Not Null,
            last_update_time Integer Not Null,
            ended Boolean Not Null,
            in_tokens Integer,
            out_tokens Integer
        );
                     ''')
    database.commit()
    print(extract_all_requests())
    print(extract_all_responses())




from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

long_description = long_description.replace('resources/demo_2x_speed.gif', 'https://github.com/jtc1246/openai-playground/blob/main/resources/demo_2x_speed.gif?raw=true')
long_description = long_description.replace('resources/chrome_extension.png', 'https://github.com/jtc1246/openai-playground/blob/main/resources/chrome_extension.png?raw=true')
long_description = long_description.replace('resources/extension_page.png', 'https://github.com/jtc1246/openai-playground/blob/main/resources/extension_page.png?raw=true')
long_description = long_description.replace('Demo:', "Demo: (if you can't see, go to [github.com/jtc1246/openai-playground](https://github.com/jtc1246/openai-playground))")
long_description = long_description.replace('(example.py)', ('(https://github.com/jtc1246/openai-playground/blob/main/example.py)'))

setup(
    name='openai-playground',
    version='1.0.5',
    author='Tiancheng Jiao',
    author_email='jtc1246@outlook.com',
    url='https://github.com/jtc1246/openai-playground',
    description='Use other openai-compatible API services in OpenAI Playground.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['openai_playground'],
    package_data={
        'openai_playground': ['*.js']
    },
    install_requires=['mySecrets', 'myHttp', 'requests'],
    python_requires='>=3.9',
    platforms=["all"],
    license='GPL-2.0 License'
)
